import logging
from typing import Any, Optional, Union

from pydantic import BaseModel, Field

from ..result import Result
from ..queries import BaseQuery


logger = logging.getLogger("opensearch_requests")


class BaseAggregation(BaseModel):
    """Base class for all aggregations."""
    size: Optional[int] = Field(None)
    """Maximum number of results to return in the aggregation."""
    name: str = Field(...)
    """Name of the aggregation."""
    filter: Optional[BaseQuery] = Field(None)
    """Filter to apply."""
    aggs: Optional[list["BaseAggregation"]] = Field(None)
    """Nested aggregations."""

    def bare(self) -> dict[str, Any]:
        result: dict[str, Any] = {}
        if self.filter:
            result["filter"] = self.filter.bare()
        if self.aggs:
            aggs: dict[str, Any] = {}
            for a in self.aggs:
                aggs[a.name] = a.bare()
            result["aggs"] = aggs
        other_fields = self.non_empty_dict()
        result.update(**other_fields)
        return result
    
    def body(self) -> dict[str, Any]:
        """Serialize aggregations to dict suitable for API request."""
        b = {
            "aggs": {
                self.name: self.bare()
            }
        }
        if self.size is not None:
            b["size"] = self.size
        return b

    def non_empty_dict(
        self, exclude: Optional[Union[set, dict]] = None
    ) -> dict[str, Any]:
        if exclude is None:
            exclude: set[str] = set()
        if isinstance(exclude, set):
            exclude = exclude.union({"size", "name"})
        elif isinstance(exclude, dict):
            exclude["size"] = True
            exclude["name"] = True

        return self.dict(
            exclude=exclude,
            exclude_none=True,
            exclude_unset=True,
        )

    def search(self, index: str, client) -> Result:
        # Make sure client object has search function as an
        # attribute that is callable
        if not (hasattr(client, "search") and callable(client.search)):
            raise RuntimeError("Wrong OpenSearch client object passed")

        body = self.body()
        logger.info(f"Aggregation body for search: {body}")

        response = client.search(
            index=index,
            body=self.body()
        )

        return Result(**response)




from enum import Enum
from typing import Any, Optional

from pydantic import BaseModel, Field


class Shards(BaseModel):
    total: int = Field(...)
    successful: int = Field(...)
    skipped: int = Field(...)
    failed: int = Field(...)


class Relation(str, Enum):
    eq = "eq"
    gte = "gte"


class Total(BaseModel):
    value: int = Field(...)
    relation: Relation = Field(...)

    class Config:
        use_enum_values = True


class Hit(BaseModel):
    index: str = Field(..., alias="_index")
    id: str = Field(..., alias="_id")
    score: str = Field(..., alias="_score")
    source: Optional[dict[str, Any]] = Field(None, alias="_source")
    fields: Optional[dict[str, list[Any]]] = Field(None)


class Hits(BaseModel):
    total: Total = Field(...)
    max_score: Optional[float] = Field(None)
    hits: list[Hit] = Field(...)


class StdDeviationBounds(BaseModel):
    upper: Optional[float] = Field(None)
    lower: Optional[float] = Field(None)
    upper_population: Optional[float] = Field(None)
    lower_population: Optional[float] = Field(None)
    upper_sampling: Optional[float] = Field(None)
    lower_sampling: Optional[float] = Field(None)


class MatrixAggregation(BaseModel):
    name: Optional[str] = Field(None)
    count: Optional[int] = Field(None)
    mean: Optional[float] = Field(None)
    variance: Optional[float] = Field(None)
    skewness: Optional[float] = Field(None)
    kurtosis: Optional[float] = Field(None)
    covariance: Optional[dict[str, float]] = Field(None)
    correlation: Optional[dict[str, float]] = Field(None)


class BucketAggregation(BaseModel):
    key: str = Field(...)
    doc_count: int = Field(...)


class Aggregation(BaseModel):
    doc_count: Optional[int] = Field(None)
    value: Optional[Any] = Field(None)
    type: Optional[Any] = Field(None)

    count: Optional[int] = Field(None)
    min: Optional[float] = Field(None)
    max: Optional[float] = Field(None)
    avg: Optional[float] = Field(None)
    sum: Optional[float] = Field(None)

    sum_of_squares: Optional[float] = Field(None)
    variance: Optional[float] = Field(None)
    variance_population: Optional[float] = Field(None)
    variance_sampling: Optional[float] = Field(None)
    std_deviation: Optional[float] = Field(None)
    std_deviation_population: Optional[float] = Field(None)
    std_deviation_sampling: Optional[float] = Field(None)
    std_deviation_bounds: Optional[StdDeviationBounds] = Field(None)
    fields: Optional[list[MatrixAggregation]] = Field(None)

    doc_count_error_upper_bound: Optional[int] = Field(None)
    sum_other_doc_count: Optional[int] = Field(None)
    buckets: Optional[list[BucketAggregation]] = Field(None)


class Result(BaseModel):
    scroll_id_: Optional[str] = Field(None, alias="_scroll_id")
    took: int = Field(...)
    timed_out: bool = Field(...)
    shards: Shards = Field(..., alias="_shards")
    hits: Optional[Hits] = Field(None)
    aggregations: Optional[dict[str, Aggregation]] = Field(None)



from typing import Optional

from pydantic import BaseModel, Field


class FieldOption(BaseModel):
    field: str = Field(...)
    """Name of the specific document field to aggregate."""


class FieldsOption(BaseModel):
    fields: list[str] = Field(...)
    """Fields to aggregate: ``["taxful_price", "base_price"]``."""


class PrecisionThresholdOption(BaseModel):
    precision_threshold: Optional[int] = Field(None)
    """Threshold below which counts are expected to be close to accurate."""


class SigmaOption(BaseModel):
    sigma: Optional[int] = Field(None)
    """Configure standard deviation for std_deviation_bounds."""


class SizeOption(BaseModel):
    """Number of top unique terms to request.
    
    Defaults to 10.
    """
    size: Optional[int] = Field(None)



import logging
from typing import Any, Optional, Union

from pydantic import BaseModel, Field

from ..result import Result


logger = logging.getLogger("opensearch_requests")


class BaseQuery(BaseModel):
    """Base class for all queries."""
    size: Optional[int] = Field(None)
    """Maximum number of results to return in the query."""

    def bare(self) -> dict[str, Any]:
        raise NotImplementedError
    
    def body(self) -> dict[str, Any]:
        """Serialize query to dict suitable for API request.
        """
        b = { "query": self.bare() }
        if self.size is not None:
            b["size"] = self.size
        return b

    def non_empty_dict(
        self, exclude: Optional[Union[set, dict]] = None
    ) -> dict[str, Any]:
        if exclude is None:
            exclude: set[str] = set()
        if isinstance(exclude, set):
            exclude = exclude.union({"size"})
        elif isinstance(exclude, dict):
            exclude["size"] = True
        return self.dict(
            exclude=exclude,
            exclude_none=True,
            exclude_unset=True,
        )

    def search(self, index: str, client) -> Result:
        # Make sure client object has search function as an
        # attribute that is callable
        if not (hasattr(client, "search") and callable(client.search)):
            raise RuntimeError("Wrong OpenSearch client object passed")

        body = self.body()
        logger.info(f"Query body for search: {body}")

        response = client.search(
            index=index,
            body=self.body()
        )

        return Result(**response)


class BaseTextQuery(BaseQuery):
    """Base class for full text queries."""
    query: str = Field(...)
    """Value to match."""



from typing import Optional

from pydantic import BaseModel, Field

from .enums import Operator, Rewrite, QueryType, ZeroTermsQuery


class AllowLeadingWildcardOption(BaseModel):
    allow_leading_wildcard: Optional[bool] = Field(None)
    """Whether * and ? are allowed as the first character of a search term.
    
    The default is true.
    """


class AnalyzeWildcardOption(BaseModel):
    analyze_wildcard: Optional[
        bool
    ] = Field(None)
    """Whether OpenSearch should attempt to analyze wildcard terms.
    
    Some analyzers do a poor job at this task, so the default is false.
    """


class AutoGenerateSynonymsPhraseQueryOption(BaseModel):
    auto_generate_synonyms_phrase_query: Optional[
        bool
    ] = Field(None)
    """Automatically generate phrase queries for multiple term synonyms.

    For example, if you have the synonym "ba, batting average" and search
    for “ba," OpenSearch searches for ba OR "batting average" when the
    option is true or ba OR (batting AND average) when the option is false.
    By default it is set to true.
    """


class BoostOption(BaseModel):
    boost: Optional[float] = Field(None)
    """Boosts the clause by the given multiplier.
    Useful for weighing clauses in compound queries. The default is 1.0.
    """


class DefaultFieldOption(BaseModel):
    default_field: Optional[str] = Field(None)


class DefaultOperatorOption(BaseModel):
    default_operator: Optional[
        Operator
    ] = Field(None)


class EnablePositionIncrimentsOption(BaseModel):
    enable_position_incriments: Optional[bool] = Field(None)
    """When true, result queries are aware of position increments.
    
    This setting is useful when the removal of stop words leaves
    an unwanted “gap" between terms. The default is true.
    """


class FieldOption(BaseModel):
    field: str = Field(...)
    """Name of the specific document field to match."""


class FieldsOption(BaseModel):
    fields: Optional[list[str]] = Field(None)
    """Fields to search: ``["title^4", "description"]``.

    The ^ lets you "boost" certain fields. Boosts are multipliers that
    weigh matches in one field more heavily than matches in other fields.

    If unspecified, defaults to the index.query.default_field setting,
    which defaults to ["*"].
    """


class FlagsOption(BaseModel):
    flags: Optional[str] = Field(None)


class FuzzinessOption(BaseModel):
    fuzziness: Optional[
        int
    ] = Field(None)
    """The number of character edits to change one word into another.

    Number of inserts, deletes, or substitutes that it takes to change
    one word to another when determining whether a term matched a value.
    For example, the distance between wined and wind is 1.
    The default, AUTO (-1 or None), chooses a value based on the length
    of each term and is a good choice for most use cases.
    """


class FuzzyMaxExpansionsOption(BaseModel):
    fuzzy_max_expansions: Optional[
        int
    ] = Field(None)
    """Limit of expanded matching terms due to fuzziness.

    Fuzzy queries "expand to" a number of matching terms that are within
    the distance specified in fuzziness.
    Then OpenSearch tries to match those terms against its indexes.
    """


class FuzzyPrefixLengthOption(BaseModel):
    fuzzy_prefix_length: Optional[
        int
    ] = Field(None)


class FuzzyTransportationsOption(BaseModel):
    fuzzy_transportations: Optional[
        bool
    ] = Field(None)
    """If true, distance between ``wind`` and ``wnid`` is 1, otherwise it's 2.

    Setting fuzzy_transpositions to true (default) adds swaps of adjacent
    characters to the insert, delete, and substitute operations of the
    fuzziness option. For example, the distance between wind and wnid
    is 1 if fuzzy_transpositions is true (swap "n" and "i") and 2 if
    it is false (delete "n", insert "n"). If fuzzy_transpositions is false,
    rewind and wnid have the same distance (2) from wind, despite the more
    human-centric opinion that wnid is an obvious typo.
    The default is a good choice for most use cases.
    """


class LenientOption(BaseModel):
    lenient: Optional[
        bool
    ] = Field(None)
    """Ignore data type mismatches.

    Setting lenient to true lets you ignore data type mismatches between
    the query and the document field. For example, a query string
    of "8.2" could match a field of type float. The default is false.
    """


class LowFreqOperatorOption(BaseModel):
    low_freq_operator: Optional[Operator] = Field(None)
    """The operator for low-frequency terms.
    
    The default is OR.
    """


class MaxExpansionsOption(BaseModel):
    max_expansions: Optional[
        int
    ] = Field(None)
    """Specifies the maximum number of terms to which the query can expand.
    
    When set, value must be a positive integer. The default is 50.
    """


class MaxDeterminedStatesOption(BaseModel):
    max_determined_states: Optional[int] = Field(None)
    """Maximum number of states for regex.

    The maximum number of “states" (a measure of complexity) that Lucene
    can create for query strings that contain regular expressions
    (e.g. "query": "/wind.+?/").
    Larger numbers allow for queries that use more memory.
    The default is 10 000.
    """


class MinimumShouldMatchOption(BaseModel):
    minimum_should_match: Optional[
        int
    ] = Field(None)
    """Number of terms that should match if OR operator is set.

    If the query string contains multiple search terms and you used
    the OR operator, the number of terms that need to match for the
    document to be considered a match.
    For example, if minimum_should_match is 2, "wind often rising"
    does not match "The Wind Rises." If minimum_should_match is 1,
    it matches.

    When set, value must be positive or negative integer, positive or
    negative percentage, or combination.
    """


class OperatorOption(BaseModel):
    operator: Optional[
        Operator
    ] = Field(None)
    """Match all terms in query (AND) or at least one (OR).

    If the query string contains multiple search terms, whether all terms
    need to match (and) or only one term needs to match (or) for a document
    to be considered a match.
    """


class PhraseSlopOption(BaseModel):
    phrase_slop: Optional[int] = Field(None)
    """Slop value for the phrase."""


class PrefixLengthOption(BaseModel):
    prefix_length: Optional[
        int
    ] = Field(None)
    """The number of leading characters that are not considered in fuzziness.
    
    The default is 0.
    """


class QuoteAnalyzerOption(BaseModel):
    quote_analyzer: Optional[str] = Field(None)


class QuoteFieldSuffixOption(BaseModel):
    quote_field_suffix: Optional[
        str
    ] = Field(None)
    """Search exact sub-field if terms are wrapped in quotes.

    This option lets you search different fields depending on whether terms
    are wrapped in quotes.
    For example, if ``quote_field_suffix`` is ``".exact"`` and you search for
    ``"lightly"`` (in quotes) in the ``title`` field, OpenSearch searches the
    ``title.exact`` field. This second field might use a different type
    (e.g. keyword rather than text) or a different analyzer.

    The default is null.
    """


class RewriteOption(BaseModel):
    rewrite: Optional[Rewrite] = Field(None)
    """Determines how OpenSearch rewrites and scores multi-term queries.
    
    The default is ``ConstantScore``.
    """


class SlopOption(BaseModel):
    slop: Optional[int] = Field(None)
    """How far words in a query can be misordered and still be a match.
    
    From the Lucene documentation: "The number of other words permitted
    between words in query phrase. For example, to switch the order
    of two words requires two moves (the first move places the words
    atop one another), so to permit re-orderings of phrases, the slop
    must be at least two. A value of zero requires an exact match."

    The default is 0. If set, value must be a postitive integer.
    """


class TieBreakerOption(BaseModel):
    tie_breaker: Optional[float] = Field(None)
    """Changes the way OpenSearch scores searches.
    
    For example, a ``type`` of ``best_fields`` typically uses the highest
    score from any one field. If you specify a ``tie_breaker`` value
    between 0.0 and 1.0, the score changes to highest
    score + ``tie_breaker`` * score for all other matching fields. If you
    specify a value of 1.0, OpenSearch adds together the scores for all
    matching fields (effectively defeating the purpose of best_fields).
    """


class TimeZoneOption(BaseModel):
    time_zone: Optional[str] = Field(None)
    """Number of hours to offset the derired time zone from UTC.

    Specifies the number of hours to offset the desired time zone
    from UTC. You need to indicate the time zone offset number
    if the query string contains a date range. For example, set
    ``time_zone: "-08:00"`` for a query with a date range such as
    "query": "wind rises release_date[2012-01-01 TO 2014-01-01]").
    
    The default time zone format used to specify number of offset
    hours is UTC.
    """


class TypeOption(BaseModel):
    type: Optional[QueryType] = Field(None)
    """Determines how OpenSearch executes the query and scores the results.

    The default is BestFields.
    """


class ZeroTermsQueryOption(BaseModel):
    zero_terms_query: Optional[
        ZeroTermsQuery
    ] = Field(None)
    """Match NONE or ALL documents, if analyzer removes all terms from query.

    If the analyzer removes all terms from a query string, whether
    to match no documents (default) or all documents.
    For example, the stop analyzer removes all terms from the string
    "an but this."
    """

    class Config:
        use_enum_values = True



from typing import Any, Union

from .base import BaseQuery


class MatchAllQuery(BaseQuery):
    """Matches and returns all documents.

    This type can be useful in testing large document sets if you need
    to return the entire set.
    """

    def bare(self) -> dict[str, Any]:
        return {
            "match_all": {},
        }



from typing import Any

from pydantic import Field

from .base import BaseTextQuery
from .options import *


class TermQuery(
    BaseTextQuery,
    FieldOption,
):
    """Search for exact 1:1 match.

    Term-level queries do not produce score. Results of term query
    will have the same score.

    Such queries are well suited for filtration.
    
    Term-level queries are not analyzed. When working with term-level
    queries on text data, only use them for fields mapped as keywords
    or arrays of strings that are explicitly ``not_analyzed``.
    """

    def bare(self) -> dict[str, Any]:
        return {
            "term": {
                self.field: self.query,
            }
        }



from typing import Any, Optional

from .base import BaseTextQuery
from .enums import Analyzer
from .options import *


class BaseFullTextQuery(BaseTextQuery):
    """Base class for full text queries."""
    analyzer: Optional[
        Analyzer
    ] = None
    """Type of analyzer to split text into structured text."""

    class Config:
        use_enum_values = True


class MatchQuery(
    BaseFullTextQuery,
    FieldOption,
    FuzzinessOption,
    FuzzyTransportationsOption,
    OperatorOption,
    MinimumShouldMatchOption,
    ZeroTermsQueryOption,
    LenientOption,
    PrefixLengthOption,
    MaxExpansionsOption,
    BoostOption,
):
    """Full-text search of a specific document field.

    The match query analyzes the provided search string and returns
    documents that match any of the string's terms.
    """
    class Config:
        use_enum_values = True

    def bare(self) -> dict[str, Any]:
        return {
            "match": {
                self.field: self.non_empty_dict(exclude={"field"}),
            }
        }


class MultiMatchQuery(
    BaseFullTextQuery,
    FieldsOption,
    TypeOption,
    OperatorOption,
    MinimumShouldMatchOption,
    TieBreakerOption,
    BoostOption,
    FuzzinessOption,
    FuzzyTransportationsOption,
    LenientOption,
    PrefixLengthOption,
    MaxExpansionsOption,
    AutoGenerateSynonymsPhraseQueryOption,
    ZeroTermsQueryOption,
    SlopOption,
):
    """Search multiple fields.

    Multi-match operation functions similarly to the match operation.
    """
    class Config:
        use_enum_values = True

    def bare(self) -> dict[str, Any]:
        return {
            "multi_match": self.non_empty_dict(),
        }


class MatchBoolPrefixQuery(
    BaseFullTextQuery,
    FieldOption,
    FuzzinessOption,
    FuzzyTransportationsOption,
    MaxExpansionsOption,
    PrefixLengthOption,
    OperatorOption,
    MinimumShouldMatchOption,
):
    """Constructs a bool query with the last term used as prefix.

    Each term except the last is used in a ``term`` query. The last term is
    used as a ``prefix`` query. 

    An important difference between the ``match_bool_prefix`` query and
    ``match_phrase_prefix`` is that the ``match_phrase_prefix`` query
    matches its terms as a phrase, but the match_bool_prefix query can
    match its terms in any position.

    Example:

        The query ``quick brown f`` could match with a field containing
        ``quick brown fox``, but it could also match ``brown fox quick``.
        It could also match a field containing the term ``quick``, the term
        ``brown`` and a term starting with ``f``, appearing in any position.

    """
    class Config:
        use_enum_values = True
    
    def bare(self) -> dict[str, Any]:
        return {
            "match_bool_prefix": {
                self.field: self.non_empty_dict(exclude={"field"}),
            }
        }


class MatchPhraseQuery(
    BaseFullTextQuery,
    FieldOption,
    SlopOption,
    ZeroTermsQueryOption,
):
    """Match documents that contain an exact phrase in a specified order.
    
    You can add flexibility to phrase matching by providing the ``slop``
    parameter.
    """
    class Config:
        use_enum_values = True
    
    def bare(self) -> dict[str, Any]:
        return {
            "match_phrase": {
                self.field: self.non_empty_dict(exclude={"field"}),
            }
        }


class MatchPhrasePrefixQuery(
    BaseFullTextQuery,
    FieldOption,
    MaxExpansionsOption,
    SlopOption,
):
    """Match a phrase in specified order.

    The documents that contain the phrase you specify will be returned.
    The last partial term in the phrase is interpreted as a prefix, so
    any documents that contain phrases that begin with the phrase and
    prefix of the last term will be returned.

    Example:

        The query ``quick brown f`` would match ``quick brown fox`` or
        ``two quick brown ferrets``, but not ``the fox is quick and brown``.

    """
    class Config:
        use_enum_values = True
    
    def bare(self) -> dict[str, Any]:
        return {
            "match_phrase_prefix": {
                self.field: self.non_empty_dict(exclude={"field"}),
            }
        }


class QueryStringQuery(
    BaseFullTextQuery,
    DefaultFieldOption,
    TypeOption,
    FuzzinessOption,
    FuzzyTransportationsOption,
    FuzzyMaxExpansionsOption,
    FuzzyPrefixLengthOption,
    MinimumShouldMatchOption,
    DefaultOperatorOption,
    LenientOption,
    BoostOption,
    AllowLeadingWildcardOption,
    EnablePositionIncrimentsOption,
    PhraseSlopOption,
    MaxDeterminedStatesOption,
    TimeZoneOption,
    QuoteFieldSuffixOption,
    QuoteAnalyzerOption,
    AnalyzeWildcardOption,
    AutoGenerateSynonymsPhraseQueryOption,
):
    """Search using a query language with operators.

    The query string query splits text based on operators and analyzes
    each individually.

    When you search using the HTTP request parameters (i.e.
    ``_search?q=wind``), OpenSearch creates a query string query.

    Example:

        ``the wind AND (rises OR rising)``

    """

    class Config:
        use_enum_values = True
    
    def bare(self) -> dict[str, Any]:
        return {
            "query_string": self.non_empty_dict(),
        }


class SimpleQueryStringQuery(
    BaseFullTextQuery,
    FieldsOption,
    FlagsOption,
    FuzzyTransportationsOption,
    FuzzyMaxExpansionsOption,
    FuzzyPrefixLengthOption,
    MinimumShouldMatchOption,
    DefaultOperatorOption,
    LenientOption,
    QuoteFieldSuffixOption,
    AnalyzeWildcardOption,
    AutoGenerateSynonymsPhraseQueryOption,
):
    """Specify multiple arguments delineated by regular expressions.

    Example:

        ``\"rises wind the\"~4 | *ising~2``

    Searches with this type will discard any invalid portions of the string.

    Syntax:
    * `+`:  `and` operator
    * `|`:  `or` operator
    * `*`:  Wildcard
    * `""`: Wrap several words (terms) into a phrase
    * `()`: Wrap a clause for precedense
    * `~n`: Set fuzziness (for example, `wnid~3`)
    * `-`:  Negate term
    """

    class Config:
        use_enum_values = True
    
    def bare(self) -> dict[str, Any]:
        return {
            "simple_query_string": self.non_empty_dict(),
        }



from typing import Any

from .base import BaseAggregation
from .options import *


class SumAggregation(
    BaseAggregation,
    FieldOption,
):
    """Returns sum value of a field."""
    # def bare(self) -> dict[str, Any]:
    #     return = super().bare()
    #     result["sum"]
    #      {
    #         "sum": self.non_empty_dict(),
    #     }


class MinAggregation(
    BaseAggregation,
    FieldOption,
):
    """Returns minimum value of a field."""
    def bare(self) -> dict[str, Any]:
        return {
            "min": self.non_empty_dict(),
        }


class MaxAggregation(
    BaseAggregation,
    FieldOption,
):
    """Returns maximum value of a field."""
    def bare(self) -> dict[str, Any]:
        return {
            "max": self.non_empty_dict(),
        }


class AvgAggregation(
    BaseAggregation,
    FieldOption,
):
    """Returns average value of a field."""
    def bare(self) -> dict[str, Any]:
        return {
            "avg": self.non_empty_dict(),
        }


class CardinalityAggregation(
    BaseAggregation,
    FieldOption,
    PrecisionThresholdOption,
):
    """Counts number of unique values in a field."""
    def bare(self) -> dict[str, Any]:
        return {
            "cardinality": self.non_empty_dict(),
        }


class ValueCountAggregation(
    BaseAggregation,
    FieldOption,
):
    """Calculates number of values that this aggregation is based on."""
    def bare(self) -> dict[str, Any]:
        return {
            "value_count": self.non_empty_dict(),
        }


class StatsAggregation(
    BaseAggregation,
    FieldOption,
):
    """Returns all basic metrics in one aggregation query.
    
    Includes: min, max, sum, avg, and value_count.
    """
    def bare(self) -> dict[str, Any]:
        return {
            "stats": self.non_empty_dict(),
        }


class ExtendedStatsAggregation(
    BaseAggregation,
    FieldOption,
    SigmaOption,
):
    """Returns extended metrics in one aggregation query.
    
    Apart from basic stats (min, max, sum, avg, value_count) also returns
    stats such as sum_of_squares, variance, and std_deviation.
    """
    def bare(self) -> dict[str, Any]:
        return {
            "extended_stats": self.non_empty_dict(),
        }


class MatrixStatsAggregation(
    BaseAggregation,
    FieldsOption,
):
    """Generates advanced stats for multiple fields."""
    def bare(self) -> dict[str, Any]:
        return {
            "matrix_stats": self.non_empty_dict(),
        }


from enum import Enum


class Analyzer(str, Enum):
    standard = "standard"
    """Parses strings into terms at word boundaries, removes punctuation.
    
    Parsing is per the Unicode text segmentation algorithm. It removes most,
    but not all, punctuation. It converts strings to lowercase.
    You can remove stop words if you turn on that option, but it does
    not remove stop words by default.
    """
    simple = "simple"
    """Splits a string into tokens on any non-letter character.

    Removes non-letter characters. Converts strings to lowercase.
    """
    whitespace = "whitespace"
    """Parses strings into terms between each whitespace."""
    stop = "stop"
    """Splits strings into tokens at each non-letter character.
    
    It also removes stop words (e.g., “but” or “this”) from strings.
    Removes non-letter characters. Converts strings to lowercase.
    """
    keyword = "keyword"
    """Outputs the entire string as one term."""
    pattern = "pattern"
    """Splits strings into terms using regular expressions.
    
    Supports converting strings to lowercase. It also supports
    removing stop words.
    """
    language = "language"
    """Provides analyzers specific to multiple languages."""
    fingerprint = "fingerprint"
    """Creates a fingerprint to use as a duplicate detector."""


class Operator(str, Enum):
    AND = "and"
    OR = "or"


class ZeroTermsQuery(str, Enum):
    NONE = "none"
    ALL = "all"


class Rewrite(str, Enum):
    ConstantScore = "constant_score"
    ScoringBoolean = "scoring_boolean"
    ConstantScoreBoolean = "constant_score_boolean"
    TopTermsN = "top_terms_N"
    TopTermsBoostN = "top_terms_bost_N"
    TopTermsBlendedFreqsN = "top_terms_blended_freqs_N"


class QueryType(str, Enum):
    BestFields = "best_fields"
    MostFields = "most_fields"
    CrossFields = "cross_fields"
    Phrase = "phrase"
    PhrasePrefix = "phrase_prefix"



from typing import Any, Optional

from pydantic import Field

from .base import BaseQuery


class BooleanQuery(BaseQuery):
    must: Optional[list[BaseQuery]] = Field(None)
    should: Optional[list[BaseQuery]] = Field(None)
    minimum_should_match: Optional[int] = Field(None, example=1)
    must_not: Optional[list[BaseQuery]] = Field(None)
    filter: Optional[list[BaseQuery]] = Field(None)

    def bare(self) -> dict[str, Any]:
        result = {
            "bool": {},
        }

        if self.must:
            result["bool"]["must"] = [q.bare() for q in self.must]
        if self.should:
            result["bool"]["should"] = [q.bare() for q in self.should]
        if self.must_not:
            result["bool"]["must_not"] = [q.bare() for q in self.must_not]
        if self.filter:
            result["bool"]["filter"] = [q.bare() for q in self.filter]
        if self.minimum_should_match is not None:
            result["bool"]["minimum_should_match"] = self.minimum_should_match

        return result



from typing import Any

from .base import BaseAggregation
from .options import FieldOption


class TermsAggregation(
    BaseAggregation,
    FieldOption,
):
    """Counts number of documents per each field value."""
    # def bare(self) -> dict[str, Any]:
    #     return {
    #         "terms": self.non_empty_dict(),
    #     }
